/*
 *  IncallDTMFtransfer.java
 *
 *  Copyright (c) 2004-2007  Brekeke Software, Inc. All rights reserved.
 *
 *  BREKEKE JTAPI SDK Sample Program
 *  
 *  Revised: May 17, 2007
 */
/******************************************************************/
/*  This sample program will receive a call at a number "1234"    */
/*  and play an anouncement, catch DTMF, transfer the call to the */
/*  DTMF number.                                                  */
/*                                                                */
/*  While calling to the destination, this program will send      */
/*  "Ringing" sound to the caller. If you want to pass the sound  */
/*  from the calling destination to the caller instead,           */
/*  please transfer right after outcall.connect.                  */
/*                                                                */
/* If you want to use REFER for this transfer, set                */                         
/*  use_refer_for_attended_transfer = true                        */
/*  in provider.properties file.                                  */   
/*                                                                */
/******************************************************************/

import java.util.*;

import javax.telephony.*;
import javax.telephony.media.*;
import javax.telephony.callcontrol.*;
import com.brekeke.jtapi.*;



public class IncallDTMFtransfer {

	static Hashtable transHash = new Hashtable(); // Hashtable for saving the call transfer statuses for each call

	public static void main(String args[]) {
		Provider myprovider = null;

		try {
			JtapiPeer peer = JtapiPeerFactory.getJtapiPeer("com.brekeke.jtapi.BrPeer"); // MUST HAVE

			/* Set your work directory of JTAPI in the second argument */
			myprovider = peer.getProvider("BrProvider; pdir=C:/JTAPI1.0/work");

			/* Add this if you need to know Register is successful or not */
			myprovider.addProviderListener(new IncallDTMFtransfer.MyProviderListener());

			/*
			 * Start provider and register this Provider's number (1234) to Brekeke
			 * SIP Server
			 */
			myprovider.start();

		} catch (Exception excp) {
			System.out.println("Can't get Provider: " + excp.toString());
			System.exit(0);
		}

		try {
			/* Choose one address from the addresses this application has */
			Terminal terminal = myprovider.getTerminal("1234");
			terminal.addCallListener(new MyCallListener());
		} catch (Exception excp) {
			System.out.println("Can't get Terminal: " + excp.toString());
			System.exit(0);
		}
	}

	/* Listener for checking Register request result */
	static class MyProviderListener implements
	com.brekeke.jtapi.BrProviderListener {

		public void eventRegistFail(BrProviderEvent event) {
			System.out.println("Register failed:" + event.getSourceUrl()+" ("+event.getStatus()+")");
		}

		public void eventRegistSuccess(BrProviderEvent event) {
			System.out.println("Register successful:" + event.getSourceUrl()+" ("+event.getStatus()+")");
		}

		public void providerInService(ProviderEvent event) {
			System.out.println("Provider bacame In Service");
		}

		public void providerEventTransmissionEnded(ProviderEvent event) {
		} /* Not supported */

		public void providerOutOfService(ProviderEvent event) {
		} /* Not supported */

		public void providerShutdown(ProviderEvent event) {
		}

	}

	static class MyCallListener implements
	javax.telephony.TerminalConnectionListener {

		public void callActive(CallEvent event) {
		}

		public void callEventTransmissionEnded(CallEvent event) {
		}

		public void callInvalid(CallEvent event) {
		}

		public void multiCallMetaMergeEnded(MetaEvent event) {
		} /* Not supported */

		public void multiCallMetaMergeStarted(MetaEvent event) {
		} /* Not supported */

		public void multiCallMetaTransferEnded(MetaEvent event) {
		} /* Not supported */

		public void multiCallMetaTransferStarted(MetaEvent event) {
		} /* Not supported */

		public void singleCallMetaProgressEnded(MetaEvent event) {
		} /* Not supported */

		public void singleCallMetaProgressStarted(MetaEvent event) {
		} /* Not supported */

		public void singleCallMetaSnapshotEnded(MetaEvent event) {
		} /* Not supported */

		public void singleCallMetaSnapshotStarted(MetaEvent event) {
		} /* Not supported */

		public void connectionAlerting(ConnectionEvent event) {
		}

		public void connectionConnected(ConnectionEvent event) {

			Connection con = event.getConnection();
			Address addr = con.getAddress();
			String addr_name = addr.getName();

			CallControlCall call = (CallControlCall) con.getCall();
			TransferStat stat = (TransferStat) transHash.get(call);

			if (stat != null) {
				System.out.println("Status=" + stat.getStatus());
			}
			System.out.println("Connection Connected:" + addr_name);

			if (addr_name.compareTo("1234") == 0) {

				if (stat != null) {
					if (stat.getStatus() <= TransferStatCode.AnsweredCall) {
						//If 1234 answers a call, call playCall method.
						try {
							playCall(call, false);
						} catch (Exception e) {
							System.out.println("Exception when playCall: " + e);
						}
					}
				} else {

					System.out.println("Can't find the call from transHash");
				}

			}

			else {
				if (stat != null) {
					//When a phone of DTMF number receives the call, transfer
					String transferadd = stat.getTransferAddr();
					if (transferadd != null) {

						if (addr_name.compareTo(transferadd) == 0 ) {
							try {
								stat.setStatus(TransferStatCode.Transfer);
								stat.setDstCall(call);
								CallControlCall origcall = (CallControlCall) stat.getOrigCall();

								Connection connectiondst = call.getConnections()[0];
								TerminalConnection[] tcs = (TerminalConnection[]) connectiondst.getTerminalConnections();
								CallControlTerminalConnection tc2 = (CallControlTerminalConnection) tcs[0];

								if (origcall != call) {
									try {
										origcall.setTransferController((TerminalConnection) stat.getOrigTC());
										call.setTransferController(tc2);
									} catch (Exception excp) {
										System.out.println("Failed to set transfer controller:"+ excp);
									}

									System.out.println("Transfer!!!");

									try {
										RingPlayer rp = stat.getRingPlayer();
										if (rp != null) {
											System.out.println("Stop playing ringing sound");

											rp.stop();
											stat.setRingPlayer(null);

										}

										call.transfer(origcall);
										stat.setStatus(TransferStatCode.Transfer_done);
									} catch (Exception excp) {

										System.out.println("Failed to transfer:"+ excp);
										excp.printStackTrace();
										try {
											playCall(origcall, true); //Transfer retry//
											call.drop();
										} catch (Exception ex) {
											System.out.println("Exception when dropping a call:"+ ex);
										}
									}

								}
							} catch (Exception ex) {
								ex.printStackTrace();
							}
						}
					}
				}
			}

		}

		public void connectionCreated(ConnectionEvent event) {
			Connection con = event.getConnection();
			Address addr = con.getAddress();
			String addr_name = addr.getName();

			System.out.println("Connection Created:" + addr_name);
		}

		public void connectionDisconnected(ConnectionEvent event) {
			Connection con = event.getConnection();
			Address addr = con.getAddress();
			String addr_name = addr.getName();

			Call call = con.getCall();

			System.out.println("Connection Disconnected:" + addr_name);

			TransferStat stat = (TransferStat) transHash.get(call);


		}

		public void connectionFailed(ConnectionEvent event) {
			Connection con = event.getConnection();
			Address addr = con.getAddress();
			String addr_name = addr.getName();

			System.out.println("Connection Failed:" + addr_name);
			CallControlCall _call = (CallControlCall) con.getCall();

			TransferStat stat = (TransferStat) transHash.get(_call);

			if (stat == null)
				System.out.println("stat = null");

			if (addr != null && stat != null) {
				System.out.println("Caller Name:" + stat.getCallerName());
				System.out.println("Stat status:" + stat.getStatus());

				if ((addr.getName()).compareTo(stat.getTransferAddr()) == 0) {

					if (stat.getStatus() == TransferStatCode.CallTransferDst) {
						try {
							CallControlCall origcall = (CallControlCall) stat.getOrigCall();
							playCall(origcall, true); //Transfer retry
						} catch (Exception e) {
							e.printStackTrace();
						}
					}

					if (stat.getStatus() > TransferStatCode.CallTransferDst) {
						if (_call != null) {
							try {
								System.out.println("Disconnect original call.");
								CallControlCall origcall = (CallControlCall) stat.getOrigCall();
								origcall.drop();

							} catch (Exception e) {
								e.printStackTrace();
							}
						}

					}
				}
			}
		}

		public void connectionInProgress(ConnectionEvent event) {
		}

		public void connectionUnknown(ConnectionEvent event) {
		}

		public void terminalConnectionActive(TerminalConnectionEvent event) {
			final TerminalConnection _tc = event.getTerminalConnection();
			Terminal term = _tc.getTerminal();

			System.out.println("Terminal Connection is active!! :" + term.getName());

		}

		public void terminalConnectionCreated(TerminalConnectionEvent event) {
			final TerminalConnection _tc = event.getTerminalConnection();
			Terminal term = _tc.getTerminal();

			System.out.println("Terminal Connection is created!! :"+ term.getName());

		}

		public void terminalConnectionDropped(TerminalConnectionEvent event) {
			TerminalConnection _tc = event.getTerminalConnection();
			Terminal term = _tc.getTerminal();
			String addr_name = term.getName();
			System.out.println("Terminal Connection is dropped:" + addr_name);

			Connection _c = _tc.getConnection();
			CallControlCall _call = (CallControlCall) _c.getCall();

			TransferStat stat = (TransferStat) transHash.get(_call);

			if (stat == null)
				System.out.println("stat = null");

			if (term != null && stat != null) {
				System.out.println("Caller Name:" + stat.getCallerName());
				System.out.println("Stat status:" + stat.getStatus());

				if ((term.getName()).compareTo(stat.getTransferAddr()) == 0
						&& stat.getStatus() >= TransferStatCode.Transfer) {
					// Transfer failed and disconnect
					if (_call != null) {
						try {
							System.out.println("Disconnect original call.");
							_call.drop();
						} catch (Exception e) {
							e.printStackTrace();
						}
					}

					CallControlCall dstcall = (CallControlCall) stat.getDstCall();
					if (dstcall != null) {
						try {
							System.out.println("Disconnect dst call.");
							dstcall.drop();
						} catch (Exception e) {
							e.printStackTrace();
						}
					}

					Address lastaddr = _call.getLastRedirectedAddress();
					if (lastaddr != null) {
						System.out.println("lastaddr=" + lastaddr.getName());
					}

				}

				if ((term.getName()).compareTo(stat.getCallerName()) == 0
						&&  stat.getStatus() == TransferStatCode.CallTransferDst) {
					// Caller disconnected before the dest answers
					if (_call != null) {
						try {
							System.out.println("Disconnect original call.");
							_call.drop();
						} catch (Exception e) {
							e.printStackTrace();
						}
					}

					CallControlCall dstcall = (CallControlCall) stat.getDstCall();
					if (dstcall != null) {
						try {
							System.out.println("Disconnect dst call.");
							dstcall.drop();
						} catch (Exception e) {
							e.printStackTrace();
						}
					}

				}

				if (stat != null) {

					if (stat.getStatus() == TransferStatCode.Transfer_done || stat.getCallerName().compareTo(addr_name)== 0) {
						transHash.remove(_call);
						_call.removeCallListener(this);
					}
				}

			}

		}

		public void terminalConnectionPassive(TerminalConnectionEvent event) {
		}

		//When 12345 receive INVITE request
		public void terminalConnectionRinging(TerminalConnectionEvent event) {

			try {
				final TerminalConnection _tc = event.getTerminalConnection();
				Terminal term = _tc.getTerminal();

				Call call = event.getCall();
				Connection[] cons;
				String callerAddr = null;
				TransferStat stat;
				String transferadd;

				System.out.println("Terminal Connection is Ringing!! :" + term.getName());

				//When 1234 receives INVITE request
				if (term.getName().compareTo("1234") == 0) {

					if (call != null) {
						cons = call.getConnections();

						if (cons.length > 0) {
							for (int i = 0; i < cons.length; ++i) {
								callerAddr = cons[i].getAddress().getName();
								if (callerAddr.compareTo("1234") != 0)
									break;
							}

						}

					}

					stat = new TransferStat(call, TransferStatCode.CallReceived);
					System.out.println("callerAddress=" + callerAddr);

					stat.setCallerName(callerAddr);

					System.out.println("Call while ringing:" + call);
					transHash.put(call, stat);

					stat.setOrigTC((CallControlTerminalConnection) _tc);

					Runnable r = new Runnable() {
						public void run() {
							try {

								_tc.answer(); // Answer the call

							} catch (Exception excp) {
								System.out.println("Failed to answer:" + excp);
							}
						};

					};
					Thread T = new Thread(r);
					T.start();

				}

				else {
					stat = (TransferStat) transHash.get(call);

					if ((stat != null)
							&& ((transferadd = stat.getTransferAddr()) != null)) {

						// Ringing the transfer destination
						if (term.getName().compareTo(transferadd) == 0) {
							try {
								System.out.println("Started Ringing play");
								RingPlayer rp = new RingPlayer(stat.getOrigCall());
								rp.start();

								stat.setRingPlayer(rp);
							} catch (Exception e) {
								System.out.println("Exception when playRinging: " + e);
							}

						}
					}

				}

			} catch (Exception excp) {
				System.out.println("Failed in terminalConnectionRinging");
			}

		}

		public void terminalConnectionUnknown(TerminalConnectionEvent event) {
		}

	}

	/** * Play an anouncement, receive DTMF, call the DTMF number ** */
	static void playCall(Call call, boolean retry) throws Exception {
		new Thread(new CallPlayer(call, retry)).start();
	}

	static class CallPlayer extends BasicMediaService implements Runnable {
		static RTC[] promptRTC = { RTC.SigDet_StopPlay };

		static final Symbol stop = SignalDetector.p_Pattern[1];
		static Hashtable patterns = new Hashtable();
		static{
			patterns.put(stop,"#");
			patterns.put(SignalDetector.p_InitialTimeout, new Integer(60000));
			patterns.put(SignalDetector.p_InterSigTimeout, new Integer(3000));
			patterns.put(SignalDetector.p_Duration, 	new Integer(60000));

		}
		static RTC[] stopRTC = { new RTC(stop, SignalDetectorConstants.rtca_Stop)};
		///////////////////

		CallControlCall call = null;

		TransferStat transstat;

		boolean retry = false;

		CallPlayer(Call call, boolean retry) {
			super((MediaProvider) call.getProvider());
			this.call = (CallControlCall) call;
			this.retry = retry;
		}

		public void run() {

			try {
				bindToCall(ConfigSpec.basicConfig, call);
				setParameters(patterns);	

				PlayerEvent playev = null;

				if (call != null) {

					transstat = (TransferStat) transHash.get(call);
					transstat.setStatus(TransferStatCode.Announce);

					if (retry == false) // Play greeting
						playev = play("/sound/exgreeting.ul", 4, promptRTC, null);
					else
						// Play retry message
						playev = play("/sound/exreentering.ul", 4, promptRTC, null);

					if (playev.getQualifier().equals(MediaConstants.q_Standard)
							|| playev.getQualifier().equals(MediaConstants.q_Stop)) {
						transstat.setStatus(TransferStatCode.AnnouncePlayed);

						System.out.println("status="+ ((TransferStat) transHash.get(call)).getStatus());

						SignalDetectorEvent sdev = null;

						while (true) {
							if (call != null) {
								try {
									transstat.setStatus(TransferStatCode.DTMFwait);
									sdev = retrieveSignals(10, null, stopRTC, null);  
									//Get DTMF signals (MAX 10 signals). If
									// caller push #, DTMF detect will be
									// stopped.


								} catch (Exception ex) {
									System.out.println("Failed to retrieve signals1:" + ex);
									System.exit(0);
								}
							}

							try {
								if (sdev.getQualifier().equals( SignalDetectorEvent.q_NumSignals)
										|| sdev.getQualifier().equals(SignalDetectorEvent.q_RTC)
										|| sdev.getQualifier().equals(SignalDetectorConstants.q_InitialTimeout)
										|| sdev.getQualifier().equals(SignalDetectorConstants.q_InterSigTimeout)
										|| sdev.getQualifier().equals(SignalDetectorConstants.q_Duration)) // New Change
								{
									String sdtmf = sdev.getSignalString();
									System.out.println("Detected DTMF =" + sdtmf);
									if(sdtmf == null) break; 

									transstat.setTransferAddr(sdtmf);
									transstat.setStatus(TransferStatCode.DTMFreceived);

									playev = null;

									// Play the number pushed
									for (int i = 0; i < sdtmf.length(); ++i) {
										if (playev == null  || playev.getQualifier().equals(
												MediaConstants.q_Standard))
											playev = play("/sound/char/" + sdtmf.charAt(i) + ".ul", 4, null, null);

										else
											break;
									}

									if (playev != null && playev.getQualifier().equals(MediaConstants.q_Standard)) {

										//Make a call from 1234 to the pushed
										// DTMF number
										Provider myProvider = call.getProvider();
										final Call outcall = myProvider.createCall();
										outcall.addCallListener(call.getCallListeners()[0]);
										final Address origaddr = myProvider.getAddress("1234");
										Terminal[] terminals = origaddr.getTerminals();
										final Terminal origterm = terminals[0];

										transstat.setDstCall(outcall);
										System.out.println("Status=" + transstat.getStatus());
										transHash.put(outcall, transstat);
										transstat.setStatus(TransferStatCode.CallTransferDst);

										Connection c[] = outcall.connect(origterm, origaddr, sdtmf);
										break;

									}
									else break; 
								} else {
									// If failed to receive DTMF
									playev = play("/sound/exreentering.ul", 4,
											promptRTC, null);
									if (playev.getQualifier().equals(
											MediaConstants.q_Standard) == false) {
										break;
									}
									// loop to try again.
								}
							} catch (Exception ex) {
								ex.printStackTrace();
								System.out.println("Failed to retrieve signals2:"
										+ ex);
							}
						}
					}
					if (isBound())
						try {
							release();
						} catch (Exception ex3) {
						}

				}
			} catch (Exception ex) {
				System.err.println(ex);
			}

		}
	}



}

